/* transform-pos.c -- use an alignment to convert a file's position
 *                    coords from one sequence to another
 *
 * This program reads a file containing position intervals which are
 * specified with respect to one sequence of a given lav or concise
 * alignment, and converts them to the corresponding positions in the
 * alignment's other sequence.  Several file formats are supported,
 * including those for exons, repeats, annotations, and underlays, or
 * any other format having lines containing two positions separated
 * by blanks.  By default the conversion is from the alignment's first
 * sequence to its second sequence, but this can be inverted using the
 * "-inverse" option.
 *
 * Lines beginning with '#' and those that do not contain a position
 * interval are left unchanged.  Lines that cannot be converted (e.g.,
 * because one or both of the positions do not align with the other
 * sequence) are flagged with "### FAILED ###" in the output, and are
 * also displayed on the screen (stderr) as a warning.
 *
 * If a position aligns with several regions in the other sequence,
 * then only the first one encountered is reported.  Thus it is best
 * if the alignment file is produced using PipMaker's "single coverage"
 * or "chaining" options.
 *
 * If the region aligns in reverse complement, the program will swap
 * the interval endpoints, and for exons and repeats files it will
 * also interchange the direction indicators ">" vs. "<" and "Right"
 * vs. "Left", respectively.  By default it will attempt to recognize
 * the file format, but this can be overridden by specifying the file
 * type on the command line.
 *
 * However, note that this program will not attempt to rearrange the
 * order of the lines.  Since in general the aligning regions might not
 * appear in the same order in the two sequences, this means the output
 * file may not be sorted.  Thus when transforming an exons file (which
 * must be sorted in increasing order for use with PipMaker), the
 * alignment file should be generated using PipMaker's "chaining" and
 * "search one strand" options; otherwise the output may have to be
 * sorted afterwards (see the sort_exons program).
 *
 * A related problem arises if an interval's two endpoints align with
 * widely-separated regions in the other sequence.  The program
 * transforms the endpoints individually, and even though it performs
 * a few basic sanity checks, there is no guarantee that the resulting
 * "interval" in the other sequence will make sense.  Again, PipMaker's
 * "chaining" and "search one strand" options can help to alleviate
 * this.  In addition, the program will print a warning if an interval's
 * endpoints are transformed using different local alignments (though
 * in that case the resulting interval might still be legitimate).
 *
 * Multiple contigs in the second sequence are not supported, since
 * the output is intended to be used as first-sequence annotation for
 * a subsequent PipMaker alignment.
 * 
 * $Revision: 1.25 $$Date: 2001/04/19 22:00:32 $
 */

#include "lav.h"
#include "feat.h"
#include "util.h"

#define SYNTAX  "position_file align_file [-inverse] " \
                  "[-exons | -repeats | -other] > output_file"
#define FAIL    "### FAILED ### "

enum { BUFSIZE = 1000 };

static void get_args (int argc, char **argv, char **posfile, char **alignfile,
                      bool *inverse, char *filetype);
static void check_exon_order (int b1, int b2, char *head, bool *validgene);

int main (int argc, char **argv) {
	char *posfile = NULL;
	char *alignfile = NULL;
	bool inverse = 0;
	char filetype = 0;   /* 'e'xons or 'r'epeats (others don't matter) */
	FILE *fp;
	char buf [BUFSIZE];
	Segment *seg;
	int size;

	/* get arguments */
	set_argv0 (argv[0]);
	get_args (argc, argv, &posfile, &alignfile, &inverse, &filetype);
	if (!posfile || !alignfile)
		fatal (usage (SYNTAX));

	/* read alignment file into Segment array */
	size = get_align (alignfile, &seg);
	if (has_multi_contig (seg))
		fatalf ("\"%s\" contains multiple contigs.", alignfile);

	/* convert position file */
	fp = ckopen (posfile, "r");
	while (fgets (buf, BUFSIZE, fp)) {
		char head [BUFSIZE], mid [BUFSIZE], *tail; /* non-addr parts of buf */
		int a1, a2;          /* original endpoints */
		int b1, b2;          /* converted endpoints */
		int n1, n2;          /* indices of segments used for conversion */
		bool gap1, gap2;     /* do endpoints align with gaps? */
		bool reverse;        /* do endpoints align in reverse complement? */
		bool validgene = 0;  /* tell check_exon_order() if gene line fails */
		char *p;

		if (strlen (buf) == BUFSIZE-1)    /* guarantees \n, 1 empty char */
			fatalf ("Line too long in \"%s\":\n%s", posfile, buf);
		if (!filetype)
			feat_guess_filetype (buf, &filetype);
		if (!feat_parse_line (buf, head, &a1, mid, &a2, &tail)) {
			fputs (buf, stdout);
			continue;
		}
		n1 = find_seg (seg, a1, inverse, NULL, 0, &gap1);
		n2 = find_seg (seg, a2, inverse, NULL, 0, &gap2);
		b1 = convert_pos (seg, n1, a1, inverse, gap1);
		b2 = convert_pos (seg, n2, a2, inverse, gap2);

		if (!b1 || !b2 ||
				(reverse = seg[n1].reverse) != seg[n2].reverse ||
				(a2 - a1) * (b2 - b1) * (reverse ? -1 : 1) < 0) {
			printf ("%s%s", FAIL, buf);
			fprintf (stderr, "%s%s", FAIL, buf);
			p = skip_ws (head);
			if (*p == '>' || *p == '<' || *p == '|')    /* gene */
				validgene = 0;
		}
		else {
			if (!same_local (seg, n1, n2)) {
				fprintf (stderr, "Warning: interval endpoints fall in " \
				                 "different local alignments:\n%s", buf);
			}
			if (reverse) {
				swap (&b1, &b2);
				feat_reverse_dir (head, tail, filetype);
				                                /* buf does have room */
			}
			if (filetype == 'e')
				check_exon_order (b1, b2, head, &validgene);
			printf ("%s%d%s%d%s", head, b1, mid, b2, tail);
		}
	}

	/* clean up */
	fclose (fp);
	free_align (seg, size);
	return 0;
}

/* Retrieve command-line arguments */
static void get_args (int argc, char **argv, char **posfile, char **alignfile,
                      bool *inverse, char *filetype) {
	int i;

	for (i = 1; i < argc; i++) {
		if (strsame (argv[i], "-inverse"))
			*inverse = 1;
		else if (strsame (argv[i], "-exons"))
			*filetype = 'e';
		else if (strsame (argv[i], "-repeats"))
			*filetype = 'r';
		else if (strsame (argv[i], "-other"))
			*filetype = 'o';
		else if (!*posfile)
			*posfile = argv[i];
		else if (!*alignfile)
			*alignfile = argv[i];
		else
			fatal (usage (SYNTAX));
	}
}

/* Print warning if the exons output file will need to be sorted */
static void check_exon_order (int b1, int b2, char *head, bool *validgene) {
	static int g1 = 0, g2 = 0;   /* current gene endpoints */
	static int e = 0;            /* watermark for exon endpoints */
	static bool warned = 0;      /* already printed warning? */
	char *p = skip_ws (head);

	if (warned)
		return;
	if (b1 > e && b1 < b2) {
		if (*p == '>' || *p == '<' || *p == '|') {    /* gene */
			if (b1 > g2) {
				g1 = b1;
				g2 = b2;
				*validgene = 1;
				return;
			}
		}
		else if (*p == '+') {    /* cds */
			if (b1 >= g1 && b2 <= (*validgene ? g2 : INT_MAX))
				return;
		}
		else {                   /* exon */
			if (b1 >= g1 && b2 <= (*validgene ? g2 : INT_MAX)) {
				e = b2;
				return;
			}
		}
	}
	fprintf (stderr, "Warning: exons output file may need to be sorted;\n"
	                 "positions are not in ascending order.\n");
	warned = 1;
}

